package org.processmining.plugins.properties.processmodel.abstractproperty;

import java.util.HashMap;
import java.util.Map;

import org.processmining.framework.util.Pair;
import org.processmining.plugins.properties.auxiliary.PerformanceOperations;
import org.processmining.plugins.properties.processmodel.gui.GUIProperty;
import org.processmining.plugins.properties.processmodel.gui.impl.GUIPropertyPointProbability;

/**
 * The point probability are tuples (value, probability) denoting the probability for a particular value
 * @author DSchunse
 *
 */
@SuppressWarnings("serial")
public abstract class PointProbability extends PropertyMapDoubleDouble{

	@SuppressWarnings("unchecked")
	public int compare(Object o1, Object o2) {
		HashMap<Double, Double> o1cast = (HashMap<Double, Double>) o1;
		HashMap<Double, Double> o2cast = (HashMap<Double, Double>) o2;
		// let's fit a 95% confidence interval
		Pair<Double, Double> o1cast95 = ninetyFivePercentConfidenceInterval(getNrOfMeasurements(o1cast), o1cast);
		Pair<Double, Double> o2cast95 = ninetyFivePercentConfidenceInterval(getNrOfMeasurements(o2cast), o2cast);
		
		if((o1cast95.getFirst() <= o2cast95.getFirst() && 
				o2cast95.getFirst() <= o1cast95.getSecond()) ||
				(o2cast95.getFirst() <= o1cast95.getFirst() && 
						o1cast95.getFirst() <= o2cast95.getSecond())){
			return 0;
		}
		else{
			return Double.compare(PerformanceOperations.mean(o1cast), PerformanceOperations.mean(o2cast));
		}
	}
	
	/**
	 * Best-effort in determining the amount of possible elements there could have been. Note this is an underapproximation,
	 * so the actual results are better.
	 * @param pointProbability
	 * @return
	 */
	private int getNrOfMeasurements(Map<Double, Double> pointProbability){
		// determine the amount of measurements in both
		int sampleSize = pointProbability.keySet().size();
		// technically, it can be that there are multiple occurrences of the same value
		double minProb = Double.MAX_VALUE;
		for(double val: pointProbability.keySet()){
			minProb = Math.min(minProb, pointProbability.get(val));
		}
		// now scale every occurrence
		int scaledSampleSize = 0;
		for(double val: pointProbability.keySet()){
			scaledSampleSize += (pointProbability.get(val) / minProb);
		}
		return Math.max(sampleSize, scaledSampleSize);
	}

	public HashMap<Double, Double> getDefaultValue(){
		HashMap<Double, Double> ret = new HashMap<Double, Double>();
		ret.put(0d, 1d);
		return ret;
	}
	
	@SuppressWarnings("unchecked")
	public double getValueDouble(Object value) {
		return PerformanceOperations.mean((HashMap<Double, Double>) value);
	}
	
	public boolean hasOrdering() {
		return true;
	}

	public boolean higherBetter() {
		return false;
	}
	
	public GUIProperty<Map<Double, Double>> getGUIProperty() throws InstantiationException, IllegalAccessException {
		return GUIPropertyPointProbability.class.newInstance();
	}
	
	public static Pair<Double, Double> ninetyFivePercentConfidenceInterval(int nrMeasurements, Map<Double, Double> pointProbability){
		Double mean = PerformanceOperations.mean(pointProbability);
		Double stdDev = PerformanceOperations.stdDev(pointProbability);
		if(nrMeasurements > 1){
			Double lowerPart = mean - giveConstant(nrMeasurements)*(stdDev / Math.sqrt(nrMeasurements));
			Double upperPart = mean + giveConstant(nrMeasurements)*(stdDev / Math.sqrt(nrMeasurements));
			return new Pair<>(lowerPart, upperPart);
		}
		else{
			// should not happen, but this is implementation technical better I guess
			return new Pair<Double, Double>(Double.MIN_VALUE, Double.MAX_VALUE);
		}
	}
	
	private static double giveConstant(int nrOfSamples){
		// nr of freedoms is 1 less than the nr of samples
		int nrOfFreedom = nrOfSamples - 1;
		if(nrOfFreedom <= 0){
			throw new UnsupportedOperationException();
		}
		// now do something with the table
		Map<Integer, Double> map95Percent = giveMap95Percent();
		if(map95Percent.keySet().contains(nrOfFreedom)){
			return map95Percent.get(nrOfFreedom);
		}
		// interpolate
		// technically, we only have a few possibilities
		// we take a linear interpolation
		if(nrOfFreedom < 40){
			return map95Percent.get(30) + ((map95Percent.get(30) - map95Percent.get(40)) / 10) * (nrOfFreedom - 30);
		}
		if(nrOfFreedom < 50){
			return map95Percent.get(40) + ((map95Percent.get(40) - map95Percent.get(50)) / 10) * (nrOfFreedom - 40);
		}
		if(nrOfFreedom < 60){
			return map95Percent.get(50) + ((map95Percent.get(50) - map95Percent.get(60)) / 10) * (nrOfFreedom - 50);
		}
		if(nrOfFreedom < 80){
			return map95Percent.get(60) + ((map95Percent.get(60) - map95Percent.get(80)) / 20) * (nrOfFreedom - 60);
		}
		if(nrOfFreedom < 100){
			return map95Percent.get(80) + ((map95Percent.get(80) - map95Percent.get(100)) / 20) * (nrOfFreedom - 80);
		}
		if(nrOfFreedom < 120){
			return map95Percent.get(100) + ((map95Percent.get(100) - map95Percent.get(120)) / 20) * (nrOfFreedom - 100);
		}
		// infinity case
		return 1.960;
	}
	
	private static Map<Integer, Double> giveMap95Percent(){
		// elements copy from http://en.wikipedia.org/wiki/Student%27s_t-distribution
		Map<Integer, Double> result = new HashMap<>();
		result.put(1, 12.71);
		result.put(2, 4.303);
		result.put(3, 3.182);
		result.put(4, 2.776);
		result.put(5, 2.571);
		result.put(6, 2.447);
		result.put(7, 2.365);
		result.put(8, 2.306);
		result.put(9, 2.262);
		result.put(10, 2.228);
		result.put(11, 2.201);
		result.put(12, 2.179);
		result.put(13, 2.16);
		result.put(14, 2.145);
		result.put(15, 2.131);
		result.put(16, 2.12);
		result.put(17, 2.11);
		result.put(18, 2.101);
		result.put(19, 2.093);
		result.put(20, 2.086);
		result.put(21, 2.08);
		result.put(22, 2.074);
		result.put(23, 2.069);
		result.put(24, 2.064);
		result.put(25, 2.06);
		result.put(26, 2.056);
		result.put(27, 2.052);
		result.put(28, 2.048);
		result.put(29, 2.045);
		result.put(30, 2.042);
		result.put(40, 2.021);
		result.put(50, 2.009);
		result.put(60, 2d);
		result.put(80, 1.99);
		result.put(100, 1.984);
		result.put(120, 1.98);
		return result;
	}
}
